import logging
import sys
import traceback
import commands
from abc import ABCMeta, abstractmethod
from cutils import console_logger, generate_config, is_metal, CommUtils
import constant
import urllib2
import re
import time,threading
import datetime
import os
import subprocess

sys.path.append(os.getcwd()+"/../comm/")
import cutils

class BaseSender(object):
    def __init__(self, logger = None):
        #print self.__class__.__module__, self.__class__.__name__

        self.__config = generate_config(constant.PLUGIN_CONFIG_PATH, self.__class__.__name__)
        self.__logger = logger if logger is not None else console_logger()
        self.init()

    def get_config(self, key):
        return self.__config[key]

    def logger(self):
        return self.__logger

    '''
        for sender plugins please override the 'send_data' function
    '''
    def init(self):
        pass

    def send_data(self, data):
        pass

class BaseCollector(object):
    __metaclass__ = ABCMeta
    def __init__(self, executor = None):
        self.__executor = executor
        self.__frequency = 60
        self.__collect_list = []
        self.__logger = executor.logger() if executor is not None else console_logger()

        self.__config = generate_config(constant.PLUGIN_CONFIG_PATH,"MetaDataServer")
        self.init()

    def set_frequency(self, freq):
        self.__frequency = freq

    def frequency(self):
        return self.__frequency

    def get_config(self, key):
        return self.__config[key]

    def logger(self):
        return self.__logger

    def collect(self):
        self.__collect_list = []
        try:
            self.do_collect()
        except Exception,e:
            error = traceback.format_exc().replace("\n","\\n")
            self.__logger.error(error)
        try:
            if self.__executor is not None:
                self.__executor.put_data(self.get_data())
        except Exception,e:
            error = traceback.format_exc().replace("\n","\\n")
            self.__logger.error(error)

    def put_data(self, data):
        if isinstance(data,list):
            self.__collect_list += data
        else:
            self.__collect_list.append(data)

    def get_data(self):
        return self.__collect_list


    def dump_data(self):
        import json
        print json.dumps(self.__collect_list, indent = 4)
    '''
        for collector plugins please override these functions below
    '''
    @abstractmethod
    def do_collect(self):
        '''
            call push_data function to fill your data
            the data format is :
            { 'sender':'sender_plugin_name', 'datas':[{}, {}, {}] }
        '''
        pass

    @abstractmethod
    def init(self):
        '''
            do some initialized job
        '''
        pass

vm_uuid = ''
ntptime_offset = 0
sn = ''
uuid_mutex = threading.Lock()
ntptime_offset_mutex = threading.Lock()
sn_mutex = threading.Lock()
metadata = ""
instance_id = ''
instance_id_mutex = threading.Lock()
app_id = ""
app_id_mutex = threading.Lock()
app_id_update_period = 3600
app_id_last_update_time = 0
barad_version = ""

class VmBaseCollector(BaseCollector):
    ntpserver = ""
    @staticmethod
    def get_vm_uuid():
        if cutils.is_minios():
	    	return ''
        global vm_uuid
        global uuid_mutex
        global metadata
        if uuid_mutex.acquire():
            if vm_uuid != '':
                uuid_mutex.release()
                return vm_uuid;
            if not cutils.is_metal():
                #get cvm uuid from dmi info
                uuid_serial = "/sys/class/dmi/id/product_serial"
                if os.path.exists(uuid_serial):
                    with open(uuid_serial,'r') as fd:
                        first_char = fd.read(1)
                        fd.seek(0)
                        if not first_char.isalnum():
                            mount_file = "/etc/uuid"
                            if os.path.exists(mount_file):
                                with open('/etc/uuid', 'r') as f:
                                    uuid = f.read().strip()
                                uuid_value = uuid.split('=')[-1].strip()
                                with open('/etc/uuid_to_serial', 'w') as f:
                                    f.write(uuid_value)
                                mount_command = "mount --bind /etc/uuid_to_serial /sys/class/dmi/id/product_serial"
                                subprocess.Popen(mount_command, shell=True)
                        vm_uuid = fd.readline().strip().lower()
                        uuid_mutex.release()
                        return vm_uuid
                uuid_path='/etc/uuid'
                if os.path.exists(uuid_path):
                    with open(uuid_path,'r') as f:
                        vm_uuid = f.read().split('=')[1].replace('"','').replace('\'','').strip()
                        uuid_mutex.release()
                        return vm_uuid
            #we have to get baremetal uuid from metadata service
            if metadata == "":
                metadata = generate_config(constant.PLUGIN_CONFIG_PATH, "MetaDataServer")
            url = metadata['metadata_url'] + 'uuid'
            response = urllib2.urlopen(url, timeout=2)
            content = response.read()
            pattern = "^[0-9a-z]{8}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{12}$"
            if re.match(pattern, content):
                vm_uuid = content
                uuid_mutex.release()
                return vm_uuid
            #set vm_uuid = ""
            vm_uuid = ""
            uuid_mutex.release()
            return vm_uuid
    def get_diskid(self, serial):
        global metadata
        if metadata == "":
            metadata = generate_config(constant.PLUGIN_CONFIG_PATH,"MetaDataServer")
        disk_id = ''
        starttime = datetime.datetime.now()
        url = ""
        if re.match("vd",serial) is None:
            url=metadata['metadata_url'] + 'bare-metal/block_disk_mapping/%s' % serial
        else:
            url=metadata['metadata_url'] + 'physical_disks_mapping/%s' % serial
        try:
            response = urllib2.urlopen(url, timeout=2)
            content = response.read()
            disk_id = content
        except Exception, e:
            disk_id = ''
        return disk_id

    def get_cbs_diskid(self,serial):
        disk_id = ''
        try:
            cmd = "ls -l  /dev/disk/by-id |sort -r -t '/' -k 3 |grep -o -E 'disk-[0-9A-Za-z]{8}'|uniq"
            output = CommUtils.ExecuteTimeoutCommand(cmd, 3).split("\n")
            disk_id = list(output)
            return disk_id
        except Exception, e:
            disk_id = ''
        return disk_id

    def get_ntp_server(self):
        self.ntpserver = generate_config(constant.PLUGIN_CONFIG_PATH,"NtpServer")
        if self.ntpserver.has_key("ntp_url"):
            return self.ntpserver['ntp_url']
        else:
            return None

    @staticmethod
    def get_barad_version():
        global barad_version
        if barad_version != "":
            return barad_version
        try:
            config_name = generate_config(constant.COMMON_CONFIG_PATH,"name")
            barad_version = config_name['agent_version'].strip().lstrip("v")
        except Exception, e:
            barad_version = ""
        return barad_version

    @staticmethod
    def get_sn():
        global sn
        global sn_mutex
        global metadata
        if sn != '':
            return sn;

        if metadata == "":
            metadata = generate_config(constant.PLUGIN_CONFIG_PATH,"MetaDataServer")

        if sn_mutex.acquire():
            if sn != '':
                sn_mutex.release()
                return sn;
            try:
                url = metadata['metadata_url'] + 'host-sn'
                response = urllib2.urlopen(url, timeout=2)
                content = response.read()
                if len(content) > 0 and len(content.splitlines()) ==1:
                    sn = content
                    sn_mutex.release()
                    return sn
            except Exception:
                sn = ''
            sn_mutex.release()
            return sn

    @staticmethod
    def get_vmip():
        try:
            import socket
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.connect(('172.0.0.1', 80))
            local_ip = sock.getsockname()[0]
        except Exception:
            import  commands
            getIpCmd = "ip route |grep -v eth0 |awk '/src/{print $NF}'"
            local_ip = commands.getoutput(getIpCmd).split("\n")[0]
        return local_ip
    
    @staticmethod
    def get_vm_instance_id():
        global instance_id
        global instance_id_mutex
        global metadata
        if instance_id != '':
            return instance_id;

        if metadata == "":
            metadata = generate_config(constant.PLUGIN_CONFIG_PATH,"MetaDataServer")

        if instance_id_mutex.acquire():
            if instance_id != '':
                instance_id_mutex.release()
                return instance_id;
            try:
                url = metadata['metadata_url'] + 'instance-id'
                response = urllib2.urlopen(url, timeout=2)
                content = response.read()
                if len(content) > 0 and len(content.splitlines()) ==1:
                    instance_id = content
                    instance_id_mutex.release()
                    return instance_id
            except Exception:
                instance_id = ''
            instance_id_mutex.release()
            return instance_id
        
    @staticmethod
    def get_vm_app_id():
        global app_id
        global app_id_mutex
        global metadata
        global app_id_update_period
        global app_id_last_update_time
        if (app_id != '') and (int(time.time()) - app_id_last_update_time < app_id_update_period):
            return app_id;

        if metadata == "":
            metadata = generate_config(constant.PLUGIN_CONFIG_PATH,"MetaDataServer")

        if app_id_mutex.acquire():
            if app_id != '':
                app_id_mutex.release()
                return app_id;
            try:
                url = metadata['metadata_url'] + 'app-id'
                response = urllib2.urlopen(url, timeout=2)
                content = response.read()
                if len(content) > 0 and len(content.splitlines()) ==1:
                    app_id = content
                    app_id_mutex.release()
                    app_id_last_update_time = (int)(time.time())
                    return app_id
            except Exception:
                app_id = ''
            app_id_mutex.release()
            return app_id


    @staticmethod
    def get_lanip():
        return VmBaseCollector.get_vmip()


    @staticmethod
    def set_ntptime_offset(offset):
        global ntptime_offset
        global ntptime_offset_mutex
        if ntptime_offset_mutex.acquire():
            ntptime_offset = offset
            ntptime_offset_mutex.release()

    @staticmethod
    def get_ntptime_offset():
        global ntptime_offset
        return ntptime_offset

class HostBaseCollector(BaseCollector):
    """
    the abstract common collector
    """
    __metaclass__ = ABCMeta
    MSG_VMM_BASE_INFO = 1
    MSG_VMM_NETWORK_INFO = 2
    MSG_VMM_STORAGE_INFO = 3
    MSG_VM_BASE_INFO = 4
    MSG_VM_NETWORK_INFO = 5
    MSG_VM_STORAGE_INFO = 6
    MSG_VM_CPU_INFO = 7
    MSG_VMM_CHILDREN_INFO = 8
    def __init__(self, executor = None):
        super(HostBaseCollector, self).__init__(executor)

    @staticmethod
    def getLocalIp():
        """Get local ip """
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.connect(('172.0.0.1', 80))
            return sock.getsockname()[0]
        except Exception, e:
            #use shell command to get local ip
            getIpCmd = "ip route |grep -v eth0 |awk '/src/{print $NF}'"
            return commands.getoutput(getIpCmd).split("\n")[0]

